@@ -1,5 +1,5 @@
-import ast
 import base64
+import collections
 import io
 import itertools
 import logging
@@ -16,7 +16,6 @@ except ImportError:
 from ..asyn import AsyncFileSystem
 from ..callbacks import _DEFAULT_CALLBACK
 from ..core import filesystem, open, split_protocol
-from ..spec import AbstractFileSystem
 from ..utils import isfilelike, merge_offset_ranges, other_paths
 
 logger = logging.getLogger("fsspec.reference")
@@ -52,24 +51,232 @@ def _protocol_groups(paths, references):
     return out
 
 
+class RefsValuesView(collections.abc.ValuesView):
+    def __iter__(self):
+        for val in self._mapping.zmetadata.values():
+            yield json.dumps(val).encode()
+        yield from self._mapping._items.values()
+        for field in self._mapping.listdir():
+            chunk_sizes = self._mapping._get_chunk_sizes(field)
+            if chunk_sizes.size == 0:
+                yield self._mapping[field + "/0"]
+                continue
+            yield from self._mapping._generate_all_records(field)
+
+
+class RefsItemsView(collections.abc.ItemsView):
+    def __iter__(self):
+        return zip(self._mapping.keys(), self._mapping.values())
+
+
+class LazyReferenceMapper(collections.abc.MutableMapping):
+    """Interface to read parquet store as if it were a standard kerchunk
+    references dict."""
+
+    # import is class level to prevent numpy dep requirement for fsspec
+    @property
+    def np(self):
+        import numpy as np
+
+        return np
+
+    @property
+    def pd(self):
+        import pandas as pd
+
+        return pd
+
+    def __init__(
+        self,
+        root,
+        fs=None,
+        cache_size=128,
+    ):
+        """
+        Parameters
+        ----------
+        root : str
+            Root of parquet store
+        fs : fsspec.AbstractFileSystem
+            fsspec filesystem object, default is local filesystem.
+        cache_size : int
+            Maximum size of LRU cache, where cache_size*record_size denotes
+            the total number of references that can be loaded in memory at once.
+        """
+        self.root = root
+        self.chunk_sizes = {}
+        self._items = {}
+        self.dirs = None
+        self.fs = fsspec.filesystem("file") if fs is None else fs
+        with self.fs.open("/".join([self.root, ".zmetadata"]), "rb") as f:
+            self._items[".zmetadata"] = f.read()
+        met = json.loads(self._items[".zmetadata"])
+        self.record_size = met["record_size"]
+        self.zmetadata = met["metadata"]
+        if self.root:
+            self.url = self.root + "/{field}/refs.{record}.parq"
+        else:
+            self.url = "{field}/refs.{record}.parq"
+
+        # Define function to open and decompress refs
+        @lru_cache(maxsize=cache_size)
+        def open_refs(field, record):
+            path = self.url.format(field=field, record=record)
+            with self.fs.open(path) as f:
+                df = self.pd.read_parquet(f, engine="fastparquet")
+            refs = {c: df[c].values for c in df.columns}
+            # Return both df and dict of views because the former is
+            # more convenient for iterating sequentially while the latter
+            # is faster for random access.
+            return df, refs
+
+        self.open_refs = open_refs
+
+    def listdir(self, basename=True):
+        if self.dirs is None:
+            dirs = [p.split("/", 1)[0] for p in self.zmetadata]
+            self.dirs = set(sorted(p for p in dirs if p and not p.startswith(".")))
+        listing = self.dirs
+        if basename:
+            listing = [os.path.basename(path) for path in listing]
+        return listing
+
+    def _load_one_key(self, key):
+        if key in self._items:
+            return self._items[key]
+        elif key in self.zmetadata:
+            return json.dumps(self.zmetadata[key]).encode()
+        elif "/" not in key:
+            raise KeyError(key)
+        field, sub_key = key.split("/")
+        # Chunk keys can be loaded from row group and cached in LRU cache
+        try:
+            record, ri, chunk_size = self._key_to_record(key)
+            if chunk_size == 0:
+                return b""
+            _, refs = self.open_refs(field, record)
+        except (ValueError, TypeError, FileNotFoundError):
+            raise KeyError(key)
+        columns = ["path", "offset", "size", "raw"]
+        selection = [refs[c][ri] if c in refs else None for c in columns]
+        raw = selection[-1]
+        if raw is not None:
+            return raw
+        data = selection[:-1]
+        if data[1:] == [0, 0]:
+            data = data[:1]
+        return data
+
+    def _key_to_record(self, key):
+        field, chunk = key.split("/")
+        chunk_sizes = self._get_chunk_sizes(field)
+        if chunk_sizes.size == 0:
+            return 0, 0, 0
+        chunk_idx = self.np.array([int(c) for c in chunk.split(".")])
+        chunk_number = self.np.ravel_multi_index(chunk_idx, chunk_sizes)
+        record = chunk_number // self.record_size
+        ri = chunk_number % self.record_size
+        return record, ri, chunk_sizes.size
+
+    def _get_chunk_sizes(self, field):
+        if field not in self.chunk_sizes:
+            zarray = self.zmetadata[f"{field}/.zarray"]
+            size_ratio = self.np.array(zarray["shape"]) / self.np.array(
+                zarray["chunks"]
+            )
+            self.chunk_sizes[field] = self.np.ceil(size_ratio).astype(int)
+        return self.chunk_sizes[field]
+
+    def _generate_record(self, field, record):
+        df, _ = self.open_refs(field, record)
+        it = df.itertuples(name=None, index=False)
+        if df.columns.size == 3:
+            # All urls
+            return (list(t) for t in it)
+        elif df.columns.size == 1:
+            # All raws
+            return (t[0] for t in it)
+        else:
+            # Mix of urls and raws
+            return (list(t[:3]) if not t[3] else t[3] for t in it)
+
+    def _generate_all_records(self, field):
+        chunk_size = self._get_chunk_sizes(field)
+        nrec = int(self.np.ceil(self.np.product(chunk_size) / self.record_size))
+        for record in range(nrec):
+            yield from self._generate_record(field, record)
+
+    def values(self):
+        return RefsValuesView(self)
+
+    def items(self):
+        return RefsItemsView(self)
+
+    def __getitem__(self, key):
+        if key in self._items:
+            val = self._items[key]
+            if val is None:
+                raise KeyError
+        if key in self.zmetadata:
+            # spec requires bytes even if we already decoded the metadata
+            return json.dumps(self.zmetadata[key]).encode()
+        return self._load_one_key(key)
+
+    def __setitem__(self, key, value):
+        self._items[key] = value
+
+    def __delitem__(self, key):
+        if key in self._items:
+            del self._items[key]
+        elif key in self.zmetadata:
+            del self.zmetadata[key]
+        else:
+            # TODO: Add method to RefFs that tracks changes in _items
+            # and updates / copies appropriate parquet ref files.
+            self._items[key] = None
+
+    def __len__(self):
+        # Caveat: This counts expected references, not actual
+        count = 0
+        for field in self.listdir():
+            if field.startswith("."):
+                count += 1
+            else:
+                chunk_sizes = self._get_chunk_sizes(field)
+                nchunks = self.np.product(chunk_sizes)
+                count += nchunks
+        count += len(self.zmetadata)  # all metadata keys
+        count += len(self._items)  # the metadata file itself
+        return count
+
+    def __iter__(self):
+        # Caveat: Note that this generates all expected keys, but does not
+        # account for reference keys that are missing.
+        yield from self.zmetadata
+        yield from self._items
+        for field in self.listdir():
+            chunk_sizes = self._get_chunk_sizes(field)
+            if chunk_sizes.size == 0:
+                yield field + "/0"
+                continue
+            inds = self.np.ndindex(*chunk_sizes)
+            for ind in inds:
+                yield field + "/" + ".".join([str(c) for c in ind])
+
+
 class ReferenceFileSystem(AsyncFileSystem):
     """View byte ranges of some other file as a file system
-
     Initial version: single file system target, which must support
     async, and must allow start and end args in _cat_file. Later versions
     may allow multiple arbitrary URLs for the targets.
-
     This FileSystem is read-only. It is designed to be used with async
     targets (for now). This FileSystem only allows whole-file access, no
     ``open``. We do not get original file details from the target FS.
-
     Configuration is by passing a dict of references at init, or a URL to
     a JSON file containing the same; this dict
     can also contain concrete data for some set of paths.
-
     Reference dict format:
     {path0: bytes_data, path1: (target_url, offset, size)}
-
     https://github.com/fsspec/kerchunk/blob/main/README.md
     """
 
@@ -89,16 +296,18 @@ class ReferenceFileSystem(AsyncFileSystem):
         simple_templates=True,
         max_gap=64_000,
         max_block=256_000_000,
+        cache_size=128,
         **kwargs,
     ):
         """
-
         Parameters
         ----------
         fo : dict or str
             The set of references to use for this instance, with a structure as above.
-            If str, will use fsspec.open, in conjunction with target_options
-            and target_protocol to open and parse JSON at this location.
+            If str referencing a JSON file, will use fsspec.open, in conjunction
+            with target_options and target_protocol to open and parse JSON at this
+            location. If a directory, then assume references are a set of parquet
+            files to be loaded lazily.
         target : str
             For any references having target_url as None, this is the default file
             target to use
@@ -140,6 +349,10 @@ class ReferenceFileSystem(AsyncFileSystem):
             number to disable merging, appropriate for local target files.
             Neighboring byte ranges will only be merged when the size of
             the aggregated range is <= ``max_block``. Default is 256MB.
+        cache_size : int
+            Maximum size of LRU cache, where cache_size*record_size denotes
+            the total number of references that can be loaded in memory at once.
+            Only used for lazily loaded references.
         kwargs : passed to parent class
         """
         super().__init__(**kwargs)
@@ -151,23 +364,27 @@ class ReferenceFileSystem(AsyncFileSystem):
         self._dircache = {}
         self.max_gap = max_gap
         self.max_block = max_block
-        if hasattr(fo, "read"):
-            text = json.load(fo)
-            text = text.decode() if isinstance(text, bytes) else text
-        elif isinstance(fo, str):
-            if target_protocol:
-                extra = {"protocol": target_protocol}
+        if isinstance(fo, str):
+            dic = dict(
+                **(ref_storage_args or target_options or {}), protocol=target_protocol
+            )
+            ref_fs, fo = fsspec.core.url_to_fs(fo, **dic)
+            if ref_fs.isfile(fo):
+                # text JSON
+                with ref_fs.open(fo, "rb") as f:
+                    logger.info("Read reference from URL %s", fo)
+                    text = json.load(f)
+                self._process_references(text, template_overrides)
             else:
-                extra = {}
-            dic = dict(**(ref_storage_args or target_options or {}), **extra)
-            # text JSON
-            with open(fo, "rb", **dic) as f:
-                logger.info("Read reference from URL %s", fo)
-                text = json.load(f)
+                # Lazy parquet refs
+                self.references = LazyReferenceMapper(
+                    fo,
+                    fs=ref_fs,
+                    cache_size=cache_size,
+                )
         else:
             # dictionaries
-            text = fo
-        self._process_references(text, template_overrides)
+            self._process_references(fo, template_overrides)
         if isinstance(fs, dict):
             self.fss = {
                 k: (
@@ -229,7 +446,7 @@ class ReferenceFileSystem(AsyncFileSystem):
             return part, None, None
 
         if len(part) == 1:
-            logger.debug(f"Reference: {path}, whole file")
+            logger.debug(f"Reference: {path}, whole file => {part}")
             url = part[0]
             start1, end1 = start, end
         else:
@@ -418,9 +635,6 @@ class ReferenceFileSystem(AsyncFileSystem):
 
     def _process_references0(self, references):
         """Make reference dict for Spec Version 0"""
-        if "zarr_consolidated_format" in references:
-            # special case for Ike prototype
-            references = _unmodel_hdf5(references)
         self.references = references
 
     def _process_references1(self, references, template_overrides=None):
@@ -631,328 +845,3 @@ class ReferenceFileSystem(AsyncFileSystem):
                 out[k] = v
         with fsspec.open(url, "wb", **storage_options) as f:
             f.write(json.dumps({"version": 1, "refs": out}).encode())
-
-
-def prefix(x):
-    if "/.z" in x or "/" not in x:
-        return "metadata", x
-    return x.split("/", 1)
-
-
-def constant_prefix(x):
-    return "metadata", x
-
-
-class DFReferenceFileSystem(AbstractFileSystem):
-    """
-    (Experimental) Parquet-based Reference Filesystem
-
-    Putative replacement or adjunct to ReferenceFileSystem with
-    additional capabilities:
-    - loads from parquet for better on-disk and in-memory space
-    - optional lazy loading by key prefix (lazy=True)
-    - multiple targets for a given key (allow_multi=True), concatenated
-      together by default, of multi_func=
-    - per-chunk processing with extra parameters stored in the parquet
-      (chunk_func=)
-
-    This implementation is not (yet) multable.
-    """
-
-    def __init__(
-        self,
-        fo,
-        target_options=None,
-        remote_protocol=None,
-        remote_options=None,
-        fs=None,
-        max_gap=64_000,
-        max_block=256_000_000,
-        parquet_kwargs=None,
-        chunk_func=None,
-        allow_multi=False,
-        multi_func=b"".join,
-        prefix_func=prefix,
-        lazy=False,
-        **kwargs,
-    ):
-        self.fo = fo
-        self.target_options = target_options or {}
-        self.max_gap = max_gap
-        self.max_block = max_block
-        self.dataframes = {}
-        self.keysets = {}
-        self.url_dict = {}
-        self.template_dict = {}
-        self.prefs = None
-        self.fss = {}
-        self.dirs = None
-        self.lazy = lazy
-        self.chunk_func = chunk_func
-        self.allow_multi = allow_multi
-        self.multi_func = multi_func
-        self.prefix_func = prefix_func if lazy else constant_prefix
-        self.pkwargs = parquet_kwargs or {}
-        if fs is not None:
-            # single remote FS
-            remote_protocol = (
-                fs.protocol[0] if isinstance(fs.protocol, tuple) else fs.protocol
-            )
-            self.fss[remote_protocol] = fs
-
-        if remote_protocol and remote_protocol not in self.fss:
-            fs = filesystem(remote_protocol, **(remote_options or {}))
-            self.fss[remote_protocol] = fs
-
-        if fs:
-            self.fss[None] = fs
-        elif self.fss:
-            self.fss[None] = iter(self.fss.values()).__next__()
-        else:
-            self.fss[None] = fsspec.filesystem(
-                remote_protocol, **(remote_options or {})
-            )
-
-        super().__init__(**kwargs)
-        self._reference_part()
-
-    def _reference_part(self, part="metadata"):
-        """Load some references from parquet
-
-        If lazy is False, this is called exactly once per instance
-
-        If lazy is true, selecting a path will determine the name of
-        the target parquet file, and the resultant columns will be
-        cached so the file need not be read again
-        """
-        import fastparquet
-
-        if part != "metadata" and part not in self.dirs:
-            raise FileNotFoundError(f"prefix {part}")
-        if part not in self.dataframes:
-            url = f"{self.fo}/{part}.parq" if self.lazy else self.fo
-            fs, path = fsspec.core.url_to_fs(url, **self.target_options)
-            pf = fastparquet.ParquetFile(path, fs=fs)
-            self.template_dict[part] = pf.key_value_metadata
-            df = pf.to_pandas()
-            thispart = {}
-            for k in df:
-                if df[k].dtype == "category" and k == "path":
-                    self.url_dict[part] = df[k].cat.categories.values
-                    thispart[k] = df[k].cat.codes.values
-                else:
-                    thispart[k] = df[k].values
-            self.dataframes[part] = thispart
-            if self.allow_multi is False:
-                self.keysets[part] = {
-                    k: i for (i, k) in enumerate(self.dataframes[part]["key"])
-                }
-            else:
-                self.keysets[part] = {}
-                for i, k in enumerate(self.dataframes[part]["key"]):
-                    self.keysets[part].setdefault(k, []).append(i)
-            if part == "metadata":
-                self.dirs = {
-                    k.rsplit("/", 1)[0]
-                    for k in self.dataframes[part]["key"]
-                    if "/" in k
-                }
-                self.prefs = (
-                    ast.literal_eval(pf.key_value_metadata["prefs"])
-                    if "prefs" in pf.key_value_metadata
-                    else set()
-                )
-
-        return self.dataframes[part]
-
-    def isdir(self, path):
-        return path in self.dirs
-
-    def cat_file(self, path, start=None, end=None, **kwargs):
-        return self.cat_ranges([path], [start], [end])[0]
-
-    def cat(self, path, recursive=False, on_error="return", **kwargs):
-        paths = self.expand_path(path, recursive=recursive)
-        paths1 = [p for p in paths if not self.isdir(p)]
-        result = {
-            p: data
-            for p, data in zip(
-                paths1, self.cat_ranges(paths1, on_error=on_error, **kwargs)
-            )
-        }
-        if len(paths1) == 1 and recursive is False and "*" not in path:
-            # same as cat_file
-            return list(result.values())[0]
-        return result
-
-    def cat_ranges(self, paths, starts=None, ends=None, on_error="return", **kwargs):
-        out = []  # eventual output; initially each key contains raw bytes or None
-        proto_dict = {}  # mapping of protocol to lists of URL/start/end to fetch
-        assign_dict = {}  # how to assign the results of cat_ranges to output
-        if starts is None:
-            starts = [None] * len(paths)
-        if ends is None:
-            ends = [None] * len(paths)
-        for p, s, e in zip(paths, starts, ends):
-            thislist = []
-            out.append(thislist)
-            if self.lazy:
-                pref, p0 = self.prefix_func(p)
-                if pref in self.prefs:
-                    # reference already inlined in metadata file
-                    pref = "metadata"
-                else:
-                    # new key in the target pref file
-                    p = p0
-            else:
-                # everything is in the same file
-                pref = "metadata"
-            self._reference_part(pref)
-            inds = self.keysets[pref][p]
-            if isinstance(inds, int):
-                inds = [inds]
-            for i in inds:
-                if x := self.dataframes[pref]["raw"][i]:
-                    thislist.append(x)
-                else:
-                    # infer path - cache this?
-                    path = self.dataframes[pref]["path"][i]
-                    if pref in self.url_dict:
-                        # dict-encoded columns; actually, numpy can do
-                        # many of these at once with int fancy indexing
-                        path = self.url_dict[pref][path]
-                    # apply template: common prefix
-                    path = path.format(**self.template_dict[pref])
-
-                    prot, _ = split_protocol(path)
-                    proto_dict.setdefault(prot, [[], [], []])
-                    proto_dict[prot][0].append(path)
-                    if s is None or s >= 0:
-                        proto_dict[prot][1].append(
-                            self.dataframes[pref]["offset"][i] or s
-                        )
-                    else:
-                        # range is from end of file, which we do not know
-                        # the size of, so this can only work if there is no
-                        # merging
-                        proto_dict[prot][1].append(s)
-
-                    if e is None or e >= 0:
-                        proto_dict[prot][2].append(
-                            self.dataframes[pref]["offset"][i]
-                            + self.dataframes[pref]["size"][i]
-                            or e
-                        )
-                    else:
-                        # range is from end of file, which we do not know
-                        # the size of, so this can only work if there is no
-                        # merging
-                        proto_dict[prot][1].append(e)
-                    thislist.append(None)
-                    assign_dict.setdefault(prot, []).append(
-                        (thislist, len(thislist) - 1)
-                    )
-
-        for proto, (urls2, starts2, ends2) in proto_dict.items():
-            fs = self.fss[proto]
-
-            new_paths, new_starts, new_ends = merge_offset_ranges(
-                list(urls2),
-                list(starts2),
-                list(ends2),
-                sort=True,
-                max_gap=self.max_gap,
-                max_block=self.max_block,
-            )
-            bytes_out = fs.cat_ranges(new_paths, new_starts, new_ends)
-            if len(urls2) == len(bytes_out):
-                # we didn't do any merging
-                for (l, i), d in zip(assign_dict[proto], bytes_out):
-                    l[i] = d
-            else:
-                # unbundle from merged bytes - simple approach
-                for u, s, e, (l, i) in zip(urls2, starts2, ends2, assign_dict[proto]):
-                    if p in out:
-                        continue  # was bytes, already handled
-                    for np, ns, ne, b in zip(
-                        new_paths, new_starts, new_ends, bytes_out
-                    ):
-                        if np == u and (ns is None or ne is None):
-                            l[i] = b[s:e]
-                        elif np == u and s >= ns and e <= ne:
-                            l[i] = b[s - ns : (e - ne) or None]
-
-        out = [self.multi_func(part) for part in out]
-        return out
-
-    def find(self, path, detail=False, withdirs=False, **kwargs):
-        path = self._strip_protocol(path)
-        if path in self.dirs:
-            path = path + "/"
-        pref, p = self.prefix_func(path)
-        dirs = (
-            [
-                {"name": d, "size": 0, "type": "directory"}
-                for d in self.dirs
-                if d.startswith(path)
-            ]
-            if withdirs
-            else []
-        )
-        if pref in self.prefs:
-            pref = "metadata"
-        df = self._reference_part(pref)
-        files = [
-            {"name": k, "type": "file", "size": _size(self.dataframes["metadata"], i)}
-            for k, i in self.keysets["metadata"].items()
-            if k.startswith(path)
-        ]
-        if self.lazy and pref != "metadata":
-            files.extend(
-                [
-                    {"name": f"{pref}/{k}", "type": "file", "size": _size(df, i)}
-                    for k, i in self.keysets[pref].items()
-                    if k.startswith(p)
-                ]
-            )
-        if detail:
-            return dirs + files
-        return [k["name"] for k in dirs + files]
-
-    def ls(self, path, detail=True, **kwargs):
-        path = self._strip_protocol(path)
-        allfiles = self.find(path, detail=True, withdirs=True)
-        isdir = path in self.dirs
-        subdfiles = [
-            p for p in allfiles if p["name"].count("/") == path.count("/") + isdir
-        ]
-        if detail:
-            return subdfiles
-        return [p["name"] for p in subdfiles]
-
-    def info(self, path, **kwargs):
-        path = self._strip_protocol(path)
-
-        if path in self.dirs:
-            return {"name": path, "type": "directory", "Size": 0}
-        return self.ls(path, detail=True)[0]
-
-
-def _size(df, i):
-    if isinstance(i, int):
-        return len(df["raw"][i]) if df["raw"][i] else df["size"][i]
-    return sum(len(df["raw"][_]) if df["raw"][_] else df["size"][_] for _ in i)
-
-
-def _unmodel_hdf5(references):
-    """Special JSON format from HDF5 prototype"""
-    # see https://gist.github.com/ajelenak/80354a95b449cedea5cca508004f97a9
-    ref = {}
-    for key, value in references["metadata"].items():
-        if key.endswith(".zchunkstore"):
-            source = value.pop("source")["uri"]
-            for k, v in value.items():
-                ref[k] = (source, v["offset"], v["offset"] + v["size"])
-        else:
-            ref[key] = json.dumps(value).encode()
-    return ref
